#! /usr/bin/perl -w
###########################################################################
## (c) Marcin Kasperski, 2008
###########################################################################
## Example FICS bot code
###########################################################################

use strict;
use Net::Telnet;

###########################################################################
# Configuration
###########################################################################

our $FICS_HOST = 'freechess.org';
our $FICS_PORT = 5000;

our $FICS_USER = 'BotTutorial';
our $FICS_PASSWORD = '';

our $CONTACT_USER = '<enter-your-name-here>';

sub finger {
    my ($username) = (@_);

    return (
        "Simple example of a FICS bot. Somebody is learning.",
        "http://blog.mekk.waw.pl/archives/7-How-to-write-a-FICS-bot-part-I.html",
        "",
        "This bot is run by $CONTACT_USER",
        "",
        "Usage:",
        "    tell $username join",
       );
}

our $SAVE_FILE = "output.txt";

our $PROTECT_LOGOUT_FREQ = 45 * 60;
our $OPEN_TIMEOUT = 30;
our $LINE_WAIT_TIMEOUT = 180;

our $VERBOSE = 1;

###########################################################################
# Global variables
###########################################################################

our $telnet;
our $username;
our $last_cmd_time = 0;
our %bad_tells;
our $grabbing_finger;

###########################################################################
# Helper functions
###########################################################################

sub register_player {
    my ($player, $rating) = @_;

    print STDERR "Registering $player ($rating)\n" if $VERBOSE;

    open(F, ">> $SAVE_FILE");
    print F "$player: $rating\n";
    close(F);
}

sub cmd_run {
    my ($cmd) = @_;
    print STDERR "Running FICS command: $cmd\n" if $VERBOSE;
    $telnet->cmd($cmd);
    $last_cmd_time = time();
}

sub process_line {
    my ($line) = @_;

    $line =~ s/[\r\n ]+$//;
    $line =~ s/^[\r\n ]+//;
    return unless $line;
    return if $line =~ /^fics%$/;

    if ($grabbing_finger) {
        if ($line =~ /^Standard\s*(\d+)/) {
            register_player($grabbing_finger, $1);
            $grabbing_finger = '';
            return;
        }
        if ($line =~ /has not played any rated games/) {
            register_player($grabbing_finger, 0);
            $grabbing_finger = '';
            return;
        }
    }

    # Somebody tells something to me 
    if ( $line =~ /^([^\s()]+)(\(\S+\))* tells you: (.*)$/ ) {
        process_tell($1, $3);
        return;
    }
  
    # Got reply to finger
    if ( $line =~ /^Finger of ([^\s\(]+)/ ) {
        $grabbing_finger = $1;
        print STDERR "Started grabbing finger of $grabbing_finger\n" if $VERBOSE;
        return;
    }

    print STDERR "Ignored line: $line\n" if $VERBOSE;
}

sub ensure_alive {
    if (time - $last_cmd_time > $PROTECT_LOGOUT_FREQ) {
        cmd_run("date");
    }
}

###########################################################################
# Main functions
###########################################################################

sub process_tell {
    my ($who, $what) = @_;
    print STDERR "Got tell from $who: $what\n";

    if ($what =~ /^\s*join\s*$/i) {
        $bad_tells{$who} = 0;
        cmd_run("finger $who /s r");
    }
    else {
        if( ++ $bad_tells{$who} < 3 ) {
            cmd_run("tell $who I do not understand your command");
        }
    }
}

sub setup {

    $telnet = new Net::Telnet(
        Timeout => $OPEN_TIMEOUT,
        Binmode => 1,
        Errmode => 'die',
       );

    $telnet->open(
        Host => $FICS_HOST,
        Port => $FICS_PORT,
       );

    print STDERR "Connected to FICS\n" if $VERBOSE;

    if ( $FICS_PASSWORD ) {

        $telnet->login(Name => $FICS_USER, Password => $FICS_PASSWORD);
        $username = $FICS_USER;
        print STDERR "Successfully logged as user $FICS_USER\n" if $VERBOSE;

    }
    else {

        $telnet->waitfor(
            Match => '/login[: ]*$/i',
            Match => '/username[: ]*$/i',
            Timeout => $OPEN_TIMEOUT);

        $telnet->print($FICS_USER);

        while (1) {
            my $line = $telnet->getline(Timeout => $LINE_WAIT_TIMEOUT);
            next if $line =~ /^[\s\r\n]*$/;
            if ($line =~ /Press return to enter/) {
                $telnet->print();
                last;
            }
            if ($line =~ /("[^"]*" is a registered name|\S+ is already logged in)/) {
                die "Can not login as $FICS_USER: $1\n";
            }
            print STDERR "Ignored line: $line\n" if $VERBOSE;
        }      

        my($pre, $match) = $telnet->waitfor(
            Match => "/Starting FICS session as ([a-zA-Z0-9]+)/",
            Match => "/\\S+ is already logged in/",
            Timeout => $OPEN_TIMEOUT);
        if ( $match =~ /Starting FICS session as ([a-zA-Z0-9]+)/ ) {
            $username = $1;
        }
        else {
            die "Can not login as $FICS_USER: $match\n";
        }

        print STDERR "Successfully logged as guest $username\n" if $VERBOSE;
    }

    # Setting FINGER
    my @finger = finger($username);
    for (my $idx = 1; $idx <= 10; ++ $idx) {
        cmd_run("set $idx " .  ($finger[$idx-1] || ''));
    }

    cmd_run("iset nowrap 1");

    cmd_run("set shout 0");
    cmd_run("set cshout 0");
    cmd_run("set seek 0");
    cmd_run("set gin 0");
    cmd_run("set pin 0");
    cmd_run("set mailmess 1");

    cmd_run("- channel 1");
    cmd_run("- channel 2");
    cmd_run("- channel 50");

    print STDERR "Finished initialization\n" if $VERBOSE;
}

sub shut_down {
    $telnet->close;
}

sub main_loop {

    $telnet->errmode(sub {
                         return if $telnet->timed_out;
                         my $msg = shift;
                         die $msg;
                     });

    while (1) {
        my $line = $telnet->getline(Timeout => $LINE_WAIT_TIMEOUT);
        if ( $line ) {
            $line =~ s/[\r\n]*$//;
            $line =~ s/^[\r\n]*//;
            process_line($line);
        }

        ensure_alive();
    }
}

###########################################################################
# Main
###########################################################################

eval {
    setup();
    main_loop();
    shut_down();
};
if ($@) {
    print STDERR "Failed: $@\n";
    exit(1);
}
